import json
from ..host.plugin_pb2 import InputArgs as ApiInputArgs, ReturnVal as ApiReturnVal, StringList

class InputArgs:
    def __init__(self, headers=None, params=None, body=None):
        if headers is None:
            headers = {}
        if params is None:
            params = {}
        self.headers = headers
        self.params = params
        self.body = body

    def to_dict(self):
        return {
            "headers": self.headers,
            "params": self.params,
            "body": self.body.decode('utf-8') if self.body else None
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str):
        data = json.loads(json_str)
        body = data.get("body")
        if body is not None:
            body = body.encode('utf-8')
        return cls(
            headers=data.get("headers"),
            params=data.get("params"),
            body=body
        )

    @classmethod
    def from_dict(cls, data):
        body = data.get("body")
        if body is not None:
            body = body.encode('utf-8')
        return cls(
            headers=data.get("headers"),
            params=data.get("params"),
            body=body
        )

    @classmethod
    def from_grpc(cls, grpc_obj: ApiInputArgs):
        headers = {}
        # for key, value in grpc_obj.headers.items():
        #     headers[key] = value
        #
        params = {}
        # for key, value in grpc_obj.params.items():
        #     params[key] = value

        return cls(
            headers=headers,
            params=params,
            body=grpc_obj.body
        )

class ReturnVal:
    def __init__(self, headers=None, params=None, value=None):
        if headers is None:
            headers = {}
        if params is None:
            params = {}
        self.headers = headers
        self.params = params
        self.value = value

    def to_dict(self):
        return {
            "headers": self.headers,
            "params": self.params,
            "value": self.value
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_grpc(self):
        headers = {}
        for key, values in self.headers.items():
            headers[key] = StringList(values=values)

        params = {}
        for key, values in self.params.items():
            params[key] = StringList(values=values)

        return ApiReturnVal(
            headers=headers,
            params=params,
            value=self.value
        )

class PluginEndpoint:
    def __init__(self, function, path, action, uses_auth, expected_args=None, expected_body=None, expected_output=None):
        self.function = function
        self.function_name = function.__name__
        self.path = path
        self.action = action
        self.uses_auth = uses_auth
        self.expected_args = expected_args
        self.expected_body = expected_body
        self.expected_output = expected_output

    def to_dict(self):
        return {
            "function_name": self.function_name,
            "path": self.path,
            "action": self.action,
            "uses_auth": self.uses_auth,
            "expected_args": self.expected_args,
            "expected_body": self.expected_body,
            "expected_output": self.expected_output
        }

    def to_json(self):
        return json.dumps(self, default=lambda o: {k: v for k, v in o.__dict__.items() if v is not None and not callable(v)}, indent=4)


# default_host.py
import os
import sys
import json
import grpc
import traceback
import dtac_tools.host.plugin_pb2_grpc

from grpc import ssl_server_credentials
from urllib.parse import quote
from concurrent import futures
from .helpers.debug_sender import DebugSender
from .helpers.encryptor import RpcEncryptor
from .plugin_host import PluginHost
from .helpers.network import get_unused_tcp_port
from ..plugins.types import InputArgs, ReturnVal
from .plugin_pb2 import PluginRequest, PluginResponse, RegisterReply, PluginEndpoint


class DefaultPluginHost(PluginHost):
    def __init__(self, plugin, debug=False, debug_port=5678):
        self.plugin = plugin
        self.rpc_proto = "grpc"
        self.proto = "tcp"
        self.ip = "127.0.0.1"
        self.interface_version = 'plug_api_1.0'
        self.port = None
        self.route_map = {}
        self.encryptor = RpcEncryptor.new_encryptor()
        if debug:
            self.debug_sender = DebugSender(debug_port)
            self.debug(f"PluginHost debugging enabled on port {debug_port}")
            self.debug(f"Plugin Name: {self.plugin.name()}")

    def debug(self, message):
        if self.debug_sender is not None:
            self.debug_sender.write(message + "\n")

    def Register(self, request, context):
        try :
            params = {}
            if request.config is not None:
                self.debug(f"config: {request.config}")
                params["config"] = request.config

            if request.default_secure is not None:
                self.debug(f"default_secure: {request.default_secure}")
                params["default_secure"] = request.default_secure

            response = self.plugin.register(params)

            # build the route map
            for endpoint in response:
                self.route_map[endpoint.path] = endpoint

            return_eps = []
            for endpoint in response:
                api_ep = PluginEndpoint(
                    path = endpoint.path,
                    action = endpoint.action,
                    uses_auth = endpoint.uses_auth,
                    expected_args = "",
                    expected_body = "",
                    expected_output = "",
                )
                return_eps.append(api_ep)

            self.debug(f"response: {response}")
            return_val = RegisterReply(
                endpoints = return_eps,
            )
            return return_val
        except Exception as ex:
            self.debug(f"Exception: {ex}")
            self.debug(traceback.format_exc())


    def Call(self, request: PluginRequest, context):
        try:
            method = request.method
            request_args = request.input_args
            self.debug(f"method: {method}")
            self.debug(f"request_args: {request_args}")
            input = InputArgs.from_grpc(request_args)
            self.debug(f"input: {input}")
            key = method.replace(f"{self.plugin.root_path()}/", "")
            self.debug(f"key: {key}")
            function = self.route_map[key].function
            ret = function(input)
            self.debug(f"output: {ret}")

            return PluginResponse(
                id = 1234,
                return_val = ReturnVal(
                    value = ret.to_json(),
                ),
                error = "",
            )

        except Exception as ex:
            self.debug(f"Exception: {ex}")
            self.debug(traceback.format_exc())

    def handle_request(self, data):
        try:
            request = json.loads(data)
            self.debug(f"request: {request}")
            if request["method"] == f"{self.plugin.name()}.Register":
                self.debug(f"Calling {request['method']} - Params: {request['params']}")
                endpoints = self.plugin.register(request["params"])
                self.debug(f"endpoints: {endpoints}")
                for endpoint in endpoints:
                    self.route_map[endpoint.function_name] = endpoint
                response = {
                    "id": request["id"],
                    "result": {
                        "Endpoints": [e.to_dict() for e in endpoints],
                    },
                    "error": None
                }
                output = json.dumps(response)
                self.debug(f"output: {output}")
                return output
            else:
                self.debug(f"Calling {request['method']} - Params: {request['params']}")
                key = request["method"].replace(f"{self.plugin.name()}.", "")
                if key in self.route_map:
                    f = self.route_map[key].function
                    input_args = InputArgs.from_dict(request["params"][0])
                    response = f(input_args)
                    output = json.dumps(response.to_dict())
                    self.debug(f"output: {output}")
                    return output
                else:
                    response = {
                        "id": request["id"],
                        "result": None,
                        "error": f"Unknown method: {request['method']}"
                    }
                    output = json.dumps(response)
                    self.debug(f"output: {output}")
                    return output
        except Exception as ex:
            self.debug(f"Exception: {ex}")
            self.debug(traceback.format_exc())

    def serve(self):

        env_cookie = os.getenv("DTAC_PLUGINS")
        if env_cookie is None:
            print('============================ WARNING ============================')
            print('This is a DTAC plugin and is not designed to be executed directly')
            print('Please use the DTAC agent to load this plugin')
            print('==================================================================')
            sys.exit(-1)

        self.port = get_unused_tcp_port()

        options = [
            f'enc={quote(self.encryptor.key_string())}',
        ]

        print(f"CONNECT{{{{{self.plugin.name()}:{self.plugin.root_path()}:{self.rpc_proto}:{self.proto}:{self.ip}:{self.port}:{self.interface_version}:[{','.join(options)}]}}}}")
        sys.stdout.flush()


        # Check for certificate and key files passed via ENV variables
        cert = os.getenv("DTAC_TLS_CERT")
        key = os.getenv("DTAC_TLS_KEY")


        # Create a gRPC server
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

        # Check if both certificate and key are provided
        if cert and key:
            # Convert certificate and key strings into bytes
            certificate_chain = cert.encode('utf-8')
            private_key = key.encode('utf-8')

            # Create server SSL credentials
            server_credentials = ssl_server_credentials(
                [(private_key, certificate_chain)]
            )

            # Add secure port using credentials
            server.add_secure_port(f"[::]:{self.port}", server_credentials)
        else:
            # Add insecure port if no TLS credentials are provided
            server.add_insecure_port(f"[::]:{self.port}")

        dtac_tools.host.plugin_pb2_grpc.add_PluginServiceServicer_to_server(self, server)
        server.add_insecure_port(f"[::]:{self.port}")
        server.start()
        server.wait_for_termination()

    def get_port(self):
        return self.port


import json
from pydantic import BaseModel
from typing import TypeVar, Type
from ..host.helpers.debug_sender import DebugSender

# Base Config class, this is used to help deserialize configuration options for plugins
class BaseConfig(BaseModel):
    pass


class PluginShared:
    @staticmethod
    def name() -> str:
        return "UnnamedPlugin"
    
    @staticmethod
    def root_path() -> str:
        return ""
    
    @staticmethod
    def serialize(v: object) -> str:
        return json.dumps(v)

class PluginBase(PluginShared):

    T = TypeVar("T", bound=BaseConfig)

    def enable_debugging(self):
        self.debug_sender = DebugSender(5678)
        self.debug(f"Plugin debugging enabled on port {5678}")
        self.debug(f"Plugin Name: {self.plugin.name()}")

    def debug(self, message):
        if self.debug_sender is not None:
            self.debug_sender.write(message + "\n")

    def load_config(self, config_json: str, config_class: Type[T]) -> T:
        return config_class.parse_raw(config_json)

    def register(self, args):
        raise NotImplementedError("this method must be implemented")

    def name(self) -> str:
        return self.__class__.__name__


from dtac_tools.host.plugin_pb2_grpc import PluginServiceServicer

class PluginHost(PluginServiceServicer):
    def serve(self) -> None:
        raise NotImplementedError

    def get_port(self) -> int:
        raise NotImplementedError


import os
import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import serialization

class RpcEncryptor:
    def __init__(self, symmetric_key):
        self.symmetric_key = symmetric_key

    @classmethod
    def new_encryptor(cls):
        key = os.urandom(32)  # AES-256 bit key
        return cls(key)

    def key_string(self):
        return base64.b64encode(self.symmetric_key).decode('utf-8')

    @classmethod
    def from_key_string(cls, key_string):
        key = base64.b64decode(key_string)
        return cls(key)

    def encrypt(self, data):
        nonce = os.urandom(12)
        cipher = Cipher(algorithms.AES(self.symmetric_key), modes.GCM(nonce), backend=default_backend())
        encryptor = cipher.encryptor()
        ciphertext = encryptor.update(data.encode('utf-8')) + encryptor.finalize()
        return base64.b64encode(nonce + ciphertext).decode('utf-8')

    def decrypt(self, encoded_data):
        data = base64.b64decode(encoded_data)
        nonce, ciphertext = data[:12], data[12:]
        cipher = Cipher(algorithms.AES(self.symmetric_key), modes.GCM(nonce), backend=default_backend())
        decryptor = cipher.decryptor()
        plaintext = decryptor.update(ciphertext) + decryptor.finalize()
        return plaintext.decode('utf-8')


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: plugin.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cplugin.proto\x12\x06plugin\"F\n\rPluginRequest\x12\x0e\n\x06method\x18\x01 \x01(\t\x12%\n\ninput_args\x18\x02 \x01(\x0b\x32\x11.plugin.InputArgs\"R\n\x0ePluginResponse\x12\n\n\x02id\x18\x01 \x01(\x05\x12%\n\nreturn_val\x18\x02 \x01(\x0b\x32\x11.plugin.ReturnVal\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"\x1c\n\nStringList\x12\x0e\n\x06values\x18\x01 \x03(\t\"\x80\x02\n\tInputArgs\x12/\n\x07headers\x18\x01 \x03(\x0b\x32\x1e.plugin.InputArgs.HeadersEntry\x12-\n\x06params\x18\x02 \x03(\x0b\x32\x1d.plugin.InputArgs.ParamsEntry\x12\x0c\n\x04\x62ody\x18\x03 \x01(\x0c\x1a\x42\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.plugin.StringList:\x02\x38\x01\x1a\x41\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.plugin.StringList:\x02\x38\x01\"\x81\x02\n\tReturnVal\x12/\n\x07headers\x18\x01 \x03(\x0b\x32\x1e.plugin.ReturnVal.HeadersEntry\x12-\n\x06params\x18\x02 \x03(\x0b\x32\x1d.plugin.ReturnVal.ParamsEntry\x12\r\n\x05value\x18\x03 \x01(\t\x1a\x42\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.plugin.StringList:\x02\x38\x01\x1a\x41\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.plugin.StringList:\x02\x38\x01\"6\n\x0cRegisterArgs\x12\x0e\n\x06\x63onfig\x18\x01 \x01(\t\x12\x16\n\x0e\x64\x65\x66\x61ult_secure\x18\x02 \x01(\x08\":\n\rRegisterReply\x12)\n\tendpoints\x18\x01 \x03(\x0b\x32\x16.plugin.PluginEndpoint\"\x88\x01\n\x0ePluginEndpoint\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0e\n\x06\x61\x63tion\x18\x03 \x01(\t\x12\x11\n\tuses_auth\x18\x04 \x01(\x08\x12\x15\n\rexpected_args\x18\x05 \x01(\t\x12\x15\n\rexpected_body\x18\x06 \x01(\t\x12\x17\n\x0f\x65xpected_output\x18\x07 \x01(\t2\x7f\n\rPluginService\x12\x37\n\x08Register\x12\x14.plugin.RegisterArgs\x1a\x15.plugin.RegisterReply\x12\x35\n\x04\x43\x61ll\x12\x15.plugin.PluginRequest\x1a\x16.plugin.PluginResponseBKZIgithub.com/intel-innersource/frameworks.automation.dtac.agent/api/grpc/gob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'plugin_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'ZIgithub.com/intel-innersource/frameworks.automation.dtac.agent/api/grpc/go'
  _INPUTARGS_HEADERSENTRY._options = None
  _INPUTARGS_HEADERSENTRY._serialized_options = b'8\001'
  _INPUTARGS_PARAMSENTRY._options = None
  _INPUTARGS_PARAMSENTRY._serialized_options = b'8\001'
  _RETURNVAL_HEADERSENTRY._options = None
  _RETURNVAL_HEADERSENTRY._serialized_options = b'8\001'
  _RETURNVAL_PARAMSENTRY._options = None
  _RETURNVAL_PARAMSENTRY._serialized_options = b'8\001'
  _globals['_PLUGINREQUEST']._serialized_start=24
  _globals['_PLUGINREQUEST']._serialized_end=94
  _globals['_PLUGINRESPONSE']._serialized_start=96
  _globals['_PLUGINRESPONSE']._serialized_end=178
  _globals['_STRINGLIST']._serialized_start=180
  _globals['_STRINGLIST']._serialized_end=208
  _globals['_INPUTARGS']._serialized_start=211
  _globals['_INPUTARGS']._serialized_end=467
  _globals['_INPUTARGS_HEADERSENTRY']._serialized_start=334
  _globals['_INPUTARGS_HEADERSENTRY']._serialized_end=400
  _globals['_INPUTARGS_PARAMSENTRY']._serialized_start=402
  _globals['_INPUTARGS_PARAMSENTRY']._serialized_end=467
  _globals['_RETURNVAL']._serialized_start=470
  _globals['_RETURNVAL']._serialized_end=727
  _globals['_RETURNVAL_HEADERSENTRY']._serialized_start=334
  _globals['_RETURNVAL_HEADERSENTRY']._serialized_end=400
  _globals['_RETURNVAL_PARAMSENTRY']._serialized_start=402
  _globals['_RETURNVAL_PARAMSENTRY']._serialized_end=467
  _globals['_REGISTERARGS']._serialized_start=729
  _globals['_REGISTERARGS']._serialized_end=783
  _globals['_REGISTERREPLY']._serialized_start=785
  _globals['_REGISTERREPLY']._serialized_end=843
  _globals['_PLUGINENDPOINT']._serialized_start=846
  _globals['_PLUGINENDPOINT']._serialized_end=982
  _globals['_PLUGINSERVICE']._serialized_start=984
  _globals['_PLUGINSERVICE']._serialized_end=1111
# @@protoc_insertion_point(module_scope)
